
from typing import Dict

from wintersweet.utils.base import Utils
from wintersweet.utils.metaclass import SafeSingleton

from elasticsearch import AsyncElasticsearch


class ElasticsearchManager(SafeSingleton):

    def __init__(self):

        self._config = {}
        self._es_map = {}
        self._ready = False

    @property
    def ready(self):
        return self._ready

    async def initialize(self):

        await self

    def register(self, config: Dict[str, Dict]):
        assert not self.ready, f'{self.__class__.__name__} already registered'
        self._config = config

    def __await__(self):
        if self._ready:
            return self

        for es_name, conf in self._config.items():

            es_client = AsyncElasticsearch(**conf)

            result = yield from es_client.ping()
            if result is False:
                raise Exception(f'"{es_name}" ping failed !')
            else:
                Utils.log.success(f'Elasticsearch [{es_name}] hosts:{es_client.transport.hosts} initialized')

            self._es_map[es_name] = es_client

        self._ready = True

    def get_client(self, es_name='default'):
        assert es_name in self._es_map, f'es_name: "{es_name}" not found'
        return self._es_map[es_name]


es_manager = ElasticsearchManager()




import copy
import logging

from fastapi import FastAPI

from wintersweet import project_label
from wintersweet.utils.logging import register_logger
from wintersweet.framework.conf import settings
from wintersweet.asyncs.pool.mysql import mysql_pool_manager
from wintersweet.asyncs.pool.redis import redis_pool_manager
from wintersweet.asyncs.pool.es import es_manager

DEFAULT_HEADERS = [(r'Server', project_label)]


def create_app():
    settings.initialize()
    app_config = copy.deepcopy(settings.APP_CONFIG)
    on_startup = app_config.get('on_startup', [])
    assert isinstance(on_startup, (list, tuple)), 'on_startup must be list or tuple'
    _on_startup = [*on_startup]
    app_config['on_startup'] = _on_startup
    if settings.DATABASES:
        mysql_pool_manager.register(settings.DATABASES)
        _on_startup.insert(0, mysql_pool_manager.initialize)

    if settings.REDIS_CONFIG:
        redis_pool_manager.register(settings.REDIS_CONFIG)
        _on_startup.insert(0, redis_pool_manager.initialize)

    if settings.ES_CONFIG:
        es_manager.register(settings.ES_CONFIG)
        _on_startup.insert(0, es_manager.initialize)

    app = FastAPI(**app_config)
    logging.getLogger(r'fastapi').setLevel(settings.LOGGING_CONFIG['level'].upper())

    register_logger(
        **settings.LOGGING_CONFIG,
        debug=settings.DEBUG
    )
    for middleware in settings.MIDDLEWARES:
        app.add_middleware(middleware['cls'], **middleware['options'])

    for exception_handler in settings.EXCEPTION_HANDLERS:
        app.add_exception_handler(exception_handler['cls'], exception_handler['handler'])

    return app




DEBUG = True
SECRET = f'__access_control_{DEBUG}_@5806fecc-93b3-11ec-beda-e2b55ff3da64__'

LISTEN_PORT = 8080

# process
PROCESS_NUM = 2


APP_CONFIG = {
    'debug': DEBUG,
    'routes': [],
}


LOGGING_CONFIG = {
    'level': 'debug',
    'file_path': '',
    'file_rotation': None,
    'file_retention': None,
}

# MIDDLEWARES
MIDDLEWARES = [

]

EXCEPTION_HANDLERS = [
    # {
    #     'cls': HTTPException,
    #     'handler': http_exception_handler
    # },
]

DATABASES = {
    # 'default': {
    #     'echo': True,
    #     'host': "localhost",
    #     'user': 'root',
    #     'password': "",
    #     'db': 'test',
    #     'port': 3306,
    #     'charset': r'utf8',
    #     'autocommit': True,
    #     'cursorclass': aiomysql.DictCursor,
    # }
}

REDIS_CONFIG = {
    # 'default': {
    #     'address': "redis://localhost:6379",
    #     'password': None,
    #     'db': 0,
    #     'encoding': 'utf-8',
    #     'minsize': 10
    # }
}

ES_CONFIG = {

}



import asyncio
import json
import ssl

from pip._vendor.certifi import where
from aiohttp import http, ClientRequest, ClientResponse, ClientWebSocketResponse, ClientTimeout, TCPConnector
from aiohttp.helpers import sentinel


DEFAULT_SSL_CONTEXT = ssl.create_default_context(cafile=where())


DEFAULT_DOWNLOAD_TIMEOUT = ClientTimeout(total=5 * 600)

_DEFAULT_SESSION_ARGS = None


async def get_session_args():

    global _DEFAULT_SESSION_ARGS

    if _DEFAULT_SESSION_ARGS is None:

        _DEFAULT_SESSION_ARGS = dict(
            connector=TCPConnector(
                use_dns_cache=True,
                ttl_dns_cache=10,
                ssl=DEFAULT_SSL_CONTEXT,
                limit=100,
                limit_per_host=10,
            ),
            loop=asyncio.get_event_loop(),
            cookies=None,
            headers=None,
            skip_auto_headers=None,
            auth=None,
            json_serialize=json.dumps,
            request_class=ClientRequest,
            response_class=ClientResponse,
            ws_response_class=ClientWebSocketResponse,
            version=http.HttpVersion11,
            cookie_jar=None,
            connector_owner=False,
            raise_for_status=False,
            read_timeout=sentinel,
            conn_timeout=None,
            timeout=sentinel,
            auto_decompress=True,
            trust_env=False,
            requote_redirect_url=True,
            trace_configs=None
            )

    return _DEFAULT_SESSION_ARGS





import asyncio
import struct
import time

from multiprocessing import shared_memory
from wintersweet.asyncs.task.tasks import IntervalTask
from wintersweet.asyncs.tools.circular import AsyncCircularForSecond
from wintersweet.utils.base import Utils

from .mysql import mysql_pool_manager
from .redis import redis_pool_manager
from .mongo import mongo_pool_manager


class _HealthShareMemory(shared_memory.SharedMemory):
    """健康状态进程间共享内存"""

    def __init__(self, name, create=False):
        super().__init__(name, create, size=8)

    def write(self, val: int):
        bytes_val = struct.pack('l', val)
        self.buf[:8] = bytes_val

    def read(self):
        bytes_val = self.buf[:8]
        return struct.unpack('l', bytes_val)[0]


class HealthChecker:
    """
        单进程服务健康检查器
        循环任务检查自身健康状况 认定节点不健康或者恢复健康往往具有一定滞后性
        Usage::
        >>> from wintersweet.asyncs.pool import HealthChecker

        >>> health_checker = HealthChecker()
        >>> await health_checker.initialize()
        >>> while True:
        >>>     print(health_checker.health)
        >>>     await asyncio.sleep(1)
    """

    def __init__(self, interval=10, timeout=30):
        self._health_interval = interval
        self._health_timeout = timeout
        self._last_check_time = 0
        self._check_task = IntervalTask(self._health_interval, self._check_health)

    async def initialize(self):
        self._check_task.start()
        await self._check_health()

    @property
    def health(self):
        return self._last_check_time + self._health_timeout > time.time()

    def check_list(self):
        return [
            mysql_pool_manager.check_health(),
            mongo_pool_manager.check_health(),
            redis_pool_manager.check_health()
        ]

    async def _check_health(self):
        result = all(await asyncio.gather(
            *self.check_list()
        ))
        if result:
            self._last_check_time = time.time()

        Utils.log.debug(f'check health:{result}')

    def release(self):
        self._check_task.stop()


class ProcessHealthChecker:

    """ 多进程健康状态检查器
        当前节点所有服务进程均健康才会认为节点健康，认定节点不健康或者恢复健康往往具有一定滞后性
        在使用 gunicorn为fastapi工程启用多进程服务时，可为该节点所有进程健康状况提供依据
        原理：
            本进程内循环任务检查自身健康状态
            当自身检查健康时将检查时间戳写入自身维护的共享内存
            读取其他进程维护的共享内存中的时间戳
            当所有时间戳均在超时范围内时，health返回True, 否则返回False

        Usage::
        >>> from wintersweet.asyncs.pool import ProcessHealthChecker

        >>> health_checker = ProcessHealthChecker(process_num=2)
        >>> await initialize()
        >>> while True:
        >>>     print(health_checker.health)
        >>>     await asyncio.sleep(1)
    """

    def __init__(self, process_num: int, interval=10, timeout=30, ntp=None):
        """
        :param process_num: 开启服务的进程数
        :param interval: 定时检查健康时间间隔
        :param timeout:  健康检查超时时间，当有进程汇报的时间戳超时timeout将被认为节点不健康
        :param ntp: ntp服务  带timestamp属性即可
        """
        self._ntp = ntp
        self._process_num = process_num
        self._health_interval = interval
        self._health_timeout = timeout
        self._memory = None
        self._health = False
        self._process_health_memory = {}
        self._check_task = IntervalTask(self._health_interval, self._check_health)

        self._is_init = False

    async def initialize(self):
        self._check_task.start()
        await self._check_health()

    @property
    def timestamp(self):
        if self._ntp is not None:
            return self._ntp.timestamp
        else:
            return int(time.time())

    async def _init_share_memory(self):

        for index in range(self._process_num):
            if index in self._process_health_memory:
                continue
            name = f'HEALTH_CHECKER_MEMORY_{index}'

            memory = None
            if self._memory is None:
                try:
                    memory = _HealthShareMemory(name=name, create=True)
                    self._memory = memory
                except:
                    memory = _HealthShareMemory(name=name)
            else:
                async for _ in AsyncCircularForSecond(max_times=5):
                    try:
                        memory = _HealthShareMemory(name=name)
                        break
                    except:
                        pass
            if memory:
                self._process_health_memory[index] = memory

        if len(self._process_health_memory) == self._process_num:
            self._is_init = True

    @property
    def health(self):

        return self._health

    def check_list(self):
        return [
            mysql_pool_manager.check_health(),
            mongo_pool_manager.check_health(),
            redis_pool_manager.check_health()
        ]

    async def _check_health(self):

        if not self._is_init:
            await self._init_share_memory()

        if self._memory is not None:
            result = await asyncio.gather(
                *self.check_list()
            )
            if all(result):
                self._memory.write(self.timestamp)

        if not self._is_init:
            self._health = False
        else:
            for item in self._process_health_memory.values():
                if item.read() + self._health_timeout < self.timestamp:
                    self._health = False
                    break
            else:
                self._health = True

        Utils.log.debug(f'check health:{self._health}')

    def release(self):
        self._check_task.stop()
        for item in self._process_health_memory.values():
            item.close()

        self._memory.unlink()
        self._process_health_memory.clear()
        self._memory = None
        self._health = False
        self._is_init = False



import os
import re
import xml
import sys
import pytz
import math
import time
import uuid
import types
import random
import socket
import loguru
import hashlib
import asyncio
import textwrap
import datetime
import binascii
import traceback
import itertools
import functools
import xmltodict


def install_uvloop():
    """安装uvloop
    """

    try:
        import uvloop
    except ModuleNotFoundError:
        Utils.log.warning(f'uvloop is not supported (T＿T)')
    else:
        uvloop.install()
        Utils.log.success(f'uvloop {uvloop.__version__} installed')


class Utils:

    os = os
    log = loguru.logger
    sys = sys
    math = math
    uuid = uuid
    time = time
    random = random
    hashlib = hashlib
    asyncio = asyncio
    datetime = datetime
    textwrap = textwrap
    traceback = traceback
    itertools = itertools
    functools = functools

    @staticmethod
    def package_task(func, *args, **kwargs):
        if asyncio.iscoroutinefunction(func):
            if args or kwargs:
                return Utils.functools.partial(func, *args, **kwargs)
            else:
                return func
        else:
            async def cro_func():
                return func(*args, **kwargs)

            return cro_func()

    @staticmethod
    @types.coroutine
    def wait_frame(count=10):
        """暂停指定帧数
        """
        for _ in range(max(1, count)):
            yield

    @classmethod
    def ip2int(cls, val):

        try:
            return int(binascii.hexlify(socket.inet_aton(val)), 16)
        except socket.error:
            return int(binascii.hexlify(socket.inet_pton(socket.AF_INET6, val)), 16)

    @classmethod
    def int2ip(cls, val):

        try:
            return socket.inet_ntoa(binascii.unhexlify(r'%08x' % val))
        except socket.error:
            return socket.inet_ntop(socket.AF_INET6, binascii.unhexlify(r'%032x' % val))

    @classmethod
    def time2stamp(cls, time_str, format_type=r'%Y-%m-%d %H:%M:%S', timezone=None):

        if timezone is None:
            return int(datetime.datetime.strptime(time_str, format_type).timestamp())
        else:
            return int(datetime.datetime.strptime(time_str, format_type).replace(tzinfo=pytz.timezone(timezone)).timestamp())

    @classmethod
    def stamp2time(cls, time_int=None, format_type=r'%Y-%m-%d %H:%M:%S', timezone=None):

        if time_int is None:
            time_int = cls.timestamp()

        if timezone is None:
            return time.strftime(format_type, datetime.datetime.fromtimestamp(time_int).timetuple())
        else:
            return time.strftime(format_type, datetime.datetime.fromtimestamp(time_int, pytz.timezone(timezone)).timetuple())

    @staticmethod
    def datetime2time(datetime_time, format_type=r'%Y-%m-%d %H:%M:%S'):
        return time.strftime(format_type, datetime_time.timetuple())

    @classmethod
    def time2datetime(cls, time_str, format_type=r'%Y-%m-%d %H:%M:%S', timezone=None):
        timestamp = cls.time2stamp(time_str, format_type, timezone=timezone)

        if timezone:
            return datetime.datetime.fromtimestamp(timestamp, pytz.timezone(timezone))
        return datetime.datetime.fromtimestamp(timestamp)

    @staticmethod
    def timestamp():
        return int(time.time())

    @staticmethod
    def delta_datetime(
            origin=False,
            add=True,
            **kwargs
    ):
        if add:
            datetime_time = datetime.datetime.now() + datetime.timedelta(**kwargs)
        else:
            datetime_time = datetime.datetime.now() - datetime.timedelta(**kwargs)

        if origin:
            datetime_time = datetime_time.replace(hour=0, minute=0, second=0, microsecond=0)

        return datetime_time

    @staticmethod
    def loop_time():
        """获取当前loop时钟
        """
        loop = asyncio.events.get_event_loop()

        return loop.time()

    @staticmethod
    def format_time(format=r'%Y-%m-%d %H:%M:%S'):
        return time.strftime(format, datetime.datetime.now().timetuple())

    @classmethod
    def re_match(cls, pattern, value):

        result = re.match(pattern, value)

        return result if result else None

    @classmethod
    def xml_encode(cls, dict_val, root_tag=r'root'):

        xml_doc = xml.dom.minidom.Document()

        root_node = xml_doc.createElement(root_tag)
        xml_doc.appendChild(root_node)

        def _convert(_doc, _node, _dict):

            for key, val in _dict.items():

                temp = _doc.createElement(key)

                if isinstance(val, dict):
                    _convert(_doc, temp, val)
                else:
                    temp.appendChild(_doc.createTextNode(str(val)))

                _node.appendChild(temp)

        _convert(xml_doc, root_node, dict_val)

        return xml_doc

    @classmethod
    def xml_decode(cls, val: str):

        return xmltodict.parse(val)

    @classmethod
    def split_int(cls, val, sep=r',', minsplit=0, maxsplit=-1):

        result = [int(item.strip()) for item in val.split(sep, maxsplit) if item.strip().lstrip(r'-').isdigit()]

        fill = minsplit - len(result)

        if fill > 0:
            result.extend(0 for _ in range(fill))

        return result



import asyncio
import time

import numpy
from ntplib import NTPClient

from wintersweet.asyncs.task.interfaces import TaskInterface
from wintersweet.asyncs.task.tasks import IntervalTask
from wintersweet.utils.base import Utils


class IntervalNTPClient(NTPClient, TaskInterface):

    def __init__(self, interval, host, version=2, port='ntp', timeout=5):

        super(IntervalNTPClient, self).__init__()
        self._running = False
        self._request_args = {
            'host': host,
            'version': version,
            'port': port,
            'timeout': timeout
        }
        self._offset = 0
        self._usability = False
        self._interval_task = IntervalTask(interval, self.run_task, tag=self.__class__.__name__)

    @property
    def offset(self):
        return self._offset

    @property
    def usability(self):
        """可用性标识"""
        return self._usability

    @property
    def timestamp(self):

        return time.time() + self._offset

    def start(self):
        if self._interval_task.running is False:
            self._interval_task.start()
        self._running = True

    def stop(self):
        if self._interval_task.running is True:
            self._interval_task.stop()
        self._running = False

    async def run_task(self):
        loop = asyncio.get_event_loop()
        offsets = []
        for _ in range(5):
            try:
                resp = await loop.run_in_executor(None, self.exec_request)
                offsets.append(resp.offset)
            except Exception as e:
                Utils.log.warning(f'NTP server {self._request_args["host"]} request failed: {e}')

        if offsets:
            self._offset = float(numpy.median(offsets))
            self._usability = True

            Utils.log.info(f'NTP server {self._request_args["host"]} offset {self._offset}')
        else:
            self._usability = False

    def exec_request(self):

        return super().request(**self._request_args)


class IntervalNTPClientManager:
    """多节点共建NTP，分布式高可用"""
    def __init__(self, interval, hosts: list, version=2, port='ntp', timeout=5):

        self._clients = [
            IntervalNTPClient(interval, host=host, version=version, port=port, timeout=timeout)
            for host in hosts
        ]

    @property
    def timestamp(self):
        clients_offset = [client.offset for client in self._clients if client.usability]
        timestamp = time.time()

        if clients_offset:
            timestamp += float(numpy.median(clients_offset))

        return timestamp

    async def initialize(self):

        await asyncio.gather(*[client.run_task() for client in self._clients])
        for client in self._clients:
            client.start()

    async def stop(self):
        _clients, self._clients = self._clients, []
        for client in self._clients:
            client.stop()



import asyncio
import json
import uuid
from contextvars import ContextVar

from aioredis import Redis


class BasePoolManager:
    def __init__(self):
        self._pools = {}
        self._config = {}
        self._ready = False

    def register(self, config):
        assert not self.ready, f'{self.__class__.__name__} already registered'
        self._config = config

    async def initialize(self):

        await self

    def __await__(self):

        raise NotImplementedError()

    def __repr__(self):
        return id(self)

    def __str__(self):
        return json.dumps({
            pool_name: self.pool_status(pool_name)
            for pool_name, pool in self._pools.items()
        }, indent=4, ensure_ascii=False)

    @property
    def ready(self):

        return self._ready

    def get_pool(self, pool_name='default'):
        return self._pools[pool_name]

    def pool_status(self, pool_name='default'):

        raise NotImplementedError()

    def get_client(self, pool_name='default', alone=False) -> Redis:

        return self._pools[pool_name].get_context_manager_client(alone)

    async def close(self):

        raise NotImplementedError()

    async def check_health(self):
        if self._pools:
            return all(await asyncio.gather(*[pool.health() for pool in self._pools.values()]))
        else:
            return True



import traceback

import aiomysql
import asyncio
import functools
import uuid
import weakref
from contextvars import ContextVar

from aiomysql import Pool

from aiomysql.sa import SAConnection, Engine
from aiomysql.sa.engine import _dialect
from sqlalchemy.sql.selectable import Select
from sqlalchemy.sql.dml import Insert, Update, Delete
from wintersweet.asyncs.pool.base import BasePoolManager
from wintersweet.asyncs.pool.context import _AsyncContextManager
from wintersweet.utils.base import Utils
from wintersweet.utils.metaclass import SafeSingleton


class MysqlClient:
    """MySQL客户端
    """
    def __init__(self, conn: SAConnection):
        self._conn = conn
        self._lock = asyncio.Lock()

    def __del__(self):
        if self._conn and not self._conn.closed:
            _conn, self._conn = self._conn, None
            asyncio.create_task(_conn.close())
            Utils.log.critical(f'Mysql:{self._conn} conn leakage!')

    @staticmethod
    def safestr(val):

        cls = type(val)

        if cls is str:
            val = aiomysql.escape_string(val)
        elif cls is dict:
            val = aiomysql.escape_dict(val, 'utf8')
        else:
            val = str(val)

        return val

    @property
    def conn(self):
        return self._conn

    @property
    def closed(self):
        return self._conn.closed

    def close(self):
        return self._conn.close()

    async def execute(self, query, *multiparams, **params):

        result = None

        async with self._lock:

            result = await self._conn.execute(query, *multiparams, **params)

        return result

    async def select(self, query, *multiparams, **params):

        result = []

        if not isinstance(query, Select):
            raise TypeError(r'Not sqlalchemy.sql.selectable.Select object')

        proxy = await self.execute(query, *multiparams, **params)

        if proxy is not None:

            records = await proxy.cursor.fetchall()

            if records:
                result = records

            if not proxy.closed:
                await proxy.close()

        return result

    async def find(self, query, *multiparams, **params):

        result = None

        if not isinstance(query, Select):
            raise TypeError(r'Not sqlalchemy.sql.selectable.Select object')

        proxy = await self.execute(query.limit(1), *multiparams, **params)

        if proxy is not None:

            record = await proxy.cursor.fetchone()

            if record:
                result = record

            if not proxy.closed:
                await proxy.close()

        return result

    async def count(self, query, *multiparams, **params):

        result = await self.find(query, *multiparams, **params)

        return result[r'tbl_row_count']

    async def insert(self, query, *multiparams, **params):

        result = 0

        if not isinstance(query, Insert):
            raise TypeError(r'Not sqlalchemy.sql.dml.Insert object')

        proxy = await self.execute(query, *multiparams, **params)

        if proxy is not None:

            result = self._conn.connection.insert_id()

            if not proxy.closed:
                await proxy.close()

        return result

    async def update(self, query, *multiparams, **params):

        result = 0

        if not isinstance(query, Update):
            raise TypeError(r'Not sqlalchemy.sql.dml.Update object')

        proxy = await self.execute(query, *multiparams, **params)

        if proxy is not None:

            result = proxy.rowcount

            if not proxy.closed:
                await proxy.close()

        return result

    async def delete(self, query, *multiparams, **params):

        result = 0

        if not isinstance(query, Delete):
            raise TypeError(r'Not sqlalchemy.sql.dml.Delete object')

        proxy = await self.execute(query, *multiparams, **params)

        if proxy is not None:

            result = proxy.rowcount

            if not proxy.closed:
                await proxy.close()

        return result

    def begin(self):
        return self._conn.begin()


class MysqlPool(Pool):

    def __init__(self, minsize=1, maxsize=10, echo=False, pool_recycle=-1, loop=None, **kwargs):

        super().__init__(minsize=minsize, maxsize=maxsize, echo=echo, pool_recycle=pool_recycle, loop=loop, **kwargs)

        self._name = None
        self._engine = None
        self._context = ContextVar(f'{self.__class__.__name__}_{uuid.uuid1().hex}', default=None)

    @property
    def name(self):
        return self._name

    def _echo_pool_info(self):
        free = self.maxsize - self.size + self.freesize
        if free < Utils.math.ceil(self.maxsize / 3):
            Utils.log.warning(
                f'Mysql pool not enough ({self._name}): (free:{free}/max:{self.maxsize})'
             )

    def get_context_client(self):
        ref = self._context.get()
        client = None
        if ref is not None:
            _client = ref() or None
            if _client and not _client.closed:
                client = _client
        return client

    async def get_client(self, alone=False):
        if self._closed:
            return
        self._echo_pool_info()

        client = MysqlClient(SAConnection(await self.acquire(), self._engine))
        if not alone:
            self._context.set(weakref.ref(client))

        return client

    def get_context_manager_client(self, alone=False):

        return _AsyncContextManager(self, alone=alone)

    async def init_fill_free_pool(self):
        if self.minsize > 0:
            async with self._cond:
                await self._fill_free_pool(False)

    async def health(self):
        try:
            async with self.get_context_manager_client(alone=True) as client:

                await client.execute('select version();')

            return True

        except:
            Utils.log.error(traceback.format_exc())
            return False


class MysqlPoolManager(SafeSingleton, BasePoolManager):
    """数据库连接池管理器"""

    def __init__(self):

        super(MysqlPoolManager, self).__init__()

    def __await__(self):
        if self._ready:
            return self

        for pool_name, pool_config in self._config.items():
            pool_config.setdefault('loop', asyncio.get_event_loop())
            pool = MysqlPool(**pool_config)
            yield from pool.init_fill_free_pool().__await__()

            self._pools[pool_name] = pool
            setattr(self._pools[pool_name], '_name', pool_name)
            setattr(self._pools[pool_name], '_engine', Engine(_dialect, pool))

            Utils.log.success(
                f"mysql pool {self.pool_status(pool_name)} initialized"
            )
        self._ready = True

        return self

    def pool_status(self, pool_name='default'):
        pool = self._pools[pool_name]
        user = self._config[pool_name].get('user', 'root')
        host = self._config[pool_name].get('host', 'localhost')
        db = self._config[pool_name].get('db', None)
        port = self._config[pool_name].get('port', 3306)

        return f'<[{pool_name}] {user}@{host}:{port} [db:{db}, size:[{pool.minsize}:{pool.maxsize}], free:{pool.freesize}]>'

    async def find(self, sql: Select, *args, pool_name='default', **kwargs):

        async with self.get_client(pool_name) as client:
            return await client.find(sql, *args, **kwargs)

    async def select(self, sql: Select, *args, pool_name='default', **kwargs):

        async with self.get_client(pool_name) as client:
            return await client.select(sql, *args, **kwargs)

    async def count(self, sql, *args, pool_name='default', **kwargs):

        async with self.get_client(pool_name) as client:
            return await client.count(sql, *args, **kwargs)

    async def update(self, sql: Update, *args, pool_name='default', **kwargs):

        async with self.get_client(pool_name) as client:
            return await client.update(sql, *args, **kwargs)

    async def insert(self, sql, *args, pool_name='default', **kwargs):

        async with self.get_client(pool_name) as client:
            return await client.insert(sql, *args, **kwargs)

    async def delete(self, sql: Delete, *args, pool_name='default', **kwargs):

        async with self.get_client(pool_name) as client:
            return await client.delete(sql, *args, **kwargs)

    async def execute(self, query, *multiparams, pool_name='default', **params):

        async with self.get_client(pool_name) as client:
            return await client.execute(query, *multiparams, **params)

    def close(self):
        self._ready = False
        _pools, self._pools = self._pools, None
        for pool in _pools.values():
            pool.close()


mysql_pool_manager = MysqlPoolManager()


def atomic(pool_name='default'):

    def wrapper(func):
        @functools.wraps(func)
        async def callback(*args, **kwargs):
            global mysql_pool_manager
            if not mysql_pool_manager.ready:
                await mysql_pool_manager

            async with mysql_pool_manager.get_client(pool_name) as trx:
                async with trx.begin():
                    result = await func(*args, **kwargs)
                    return result
        return callback
    return wrapper




import asyncio
import copy
import functools
import hashlib
import json
import traceback

from cachetools import TTLCache

from wintersweet.asyncs.pool.redis import redis_pool_manager
from wintersweet.utils.base import Utils


class ShareCache:

    """
    并发缓存器
        实现功能：瞬时并发执行某个cro函数，通过计算签名相同时，只执行一次，并且支持缓存到redis以及从redis取
        注意事项：1、当使用ShareCache修饰的函数，其参数需支持json序列化
                当为普通对象时因计算签名不一致将无法起到缓存作用
                2、当使用redis作为缓存介质时，将对返回值进行json序列化，因此返回值仅限基本类型，不可为普通对象
                当缓存反序列化后，原返回值中的tuple类型将映射为list类型返回
                因此，为避免前后调用返回结果不一致，请确保返回值中不包含不可序列化对象或经反序列化不产生结构和类型变化的对象
    Usage::
    >>> @ShareCache(cache_none=True, use_redis=True, ttl=30)
    >>> async def test(*args, **kwargs):
    >>>     ...
    >>>     return [1, 2]   # 建议，序列化反序列化后结构和类型不发生变化，都是list
    >>>     return 1, 2   # 不建议，序列化前是tuple，反序列化后是list，使用缓存时，将导致第一次调用返回tuple，第二次调用从缓存取会得到list

    >>> result = await test(1, 2)
    """

    def __init__(self, tag=None, cache_none=False, use_redis=True, ttl=30):
        """
        :param tag: 被装饰的函数唯一标识，如果tag为None，计算签名将使用函数名作为函数唯一标识
        :param cache_none: 是否缓存返回值None
        :param use_redis: 是否使用缓存
        :param ttl: 缓存时长，单位s
        """
        self._ttl = ttl
        self._cache_none = cache_none
        self._use_redis = use_redis
        self._futures = {}
        self._tag = tag

    def __call__(self, func):

        @functools.wraps(func)
        async def _wrapper(*args, **kwargs):

            func_sign = self._sig(func, *args, **kwargs)
            if func_sign in self._futures:
                f = asyncio.Future()
                self._futures[func_sign].append(f)
            else:
                f = asyncio.create_task(self.wrapper_func(func_sign, func, *args, **kwargs))
                if not f:
                    raise RuntimeError(f'"{func}" is not a cro function')

                self._futures[func_sign] = [f]
                f.add_done_callback(functools.partial(self._set_result, func_sign))

            return await f

        return _wrapper

    def _sig(self, func, *args, **kwargs):
        hash_str = f'{self._tag if self._tag else func.__name__}%@@%{args}%@@%{kwargs}'
        return hashlib.md5(hash_str.encode()).hexdigest()

    def _set_result(self, func_sign, _):

        if func_sign not in self._futures:
            return

        futures = self._futures.pop(func_sign)

        result = futures.pop(0).result()

        for future in futures:
            future.set_result(copy.deepcopy(result))

    async def wrapper_func(self, func_sign, func, *args, **kwargs):

        if not self._use_redis:
            return await func(*args, **kwargs)

        result = None
        async with redis_pool_manager.get_client() as client:

            _result = await client.get(func_sign)
            if _result:
                try:
                    result = json.loads(_result)['result']
                    if self._cache_none:
                        if result is None:
                            return result

                except Exception:
                    Utils.log.warning(f'{self.__class__.__name__} got unexpected result "{_result}"')
                    await client.delete(func_sign)

        if result:
            return result

        result = await func(*args, **kwargs)

        async with redis_pool_manager.get_client() as client:
            try:
                await client.set(func_sign, json.dumps({'result': result}), expire=self._ttl)
            except:
                Utils.log.error(traceback.format_exc())

        return result


class CacheWrap:
    """函数缓存器"""
    def __init__(self, maxsize=0xffff, ttl=3600):
        self._cache = TTLCache(maxsize=maxsize, ttl=ttl)

    def _sig(self, func, *args, **kwargs):
        hash_str = f'{func.__name__}%@@%{args}%@@%{kwargs}'
        return hashlib.md5(hash_str.encode()).hexdigest()

    def __call__(self, func):

        @functools.wraps(func)
        async def _wrapper(*args, **kwargs):

            func_sign = self._sig(func, *args, **kwargs)
            result = self._cache.get(func_sign, None)

            if result is None:
                result = await func(*args, **kwargs)
                self._cache[func_sign] = result

            return result

        return _wrapper



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'  # copy

from contextvars import ContextVar

from starlette.types import ASGIApp, Message, Receive, Scope, Send

from wintersweet.utils.base import Utils

REQUEST_ID_CONTEXT = ContextVar(r'request_id')


class RequestIDMiddleware:

    @staticmethod
    def get_request_id():

        request_id = REQUEST_ID_CONTEXT.get(None)

        if request_id is None:
            Utils.log.warning(r'RequestIDMiddleware is not enabled!')

        return request_id

    def __init__(self, app: ASGIApp) -> None:

        self._app = app

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:

        responder = _Response(self._app)

        await responder(scope, receive, send)


class _Response:

    def __init__(self, app: ASGIApp):

        self._app = app
        self._send = None

        self._request_id = Utils.uuid.uuid1().hex

        REQUEST_ID_CONTEXT.set(self._request_id)

    async def __call__(self, scope: Scope, receive: Receive, send: Send):

        self._send = send

        await self._app(scope, receive, self.send)

    async def send(self, message: Message) -> None:

        message.setdefault(r'headers', [])

        message[r'headers'].append((b'x-timestamp', str(Utils.timestamp()).encode(r'latin-1')))
        message[r'headers'].append((b'x-request-id', self._request_id.encode(r'latin-1')))

        await self._send(message)




import aiohttp


def _build_form_data(data, files):
    """
    :param data:
    :param files: {'file1': ('file1.txt', open('file1.txt')), 'file2': ('file2.txt', open('file2.txt'))}
    :return: FormData
    """
    form = aiohttp.FormData()
    if data:
        for k, v in data.items():
            form.add_field(k, str(v))

    if files:
        for name, file in files:
            form.add_field(name, value=file[1], filename=file[0])

    return form


@aiohttp.streamer
def _file_sender(writer, file_path):
    with open(file_path, 'rb') as f:
        chunk = f.read(65535)
        while chunk:
            yield from writer.write(chunk)
            chunk = f.read(65535)







import aioredis
import asyncio
import json
import time
import traceback
import uuid
import weakref

from contextvars import ContextVar
from aioredis import Redis, ConnectionsPool

from wintersweet.asyncs.pool.base import BasePoolManager
from wintersweet.asyncs.pool.context import _AsyncContextManager
from wintersweet.asyncs.task.tasks import IntervalTask
from wintersweet.asyncs.tools.circular import AsyncCircularForTimeout, AsyncCircularForSecond
from wintersweet.utils.base import Utils
from wintersweet.utils.errors import catch_error
from wintersweet.utils.metaclass import SafeSingleton


class RedisPool(ConnectionsPool):

    def __init__(self, *args, **kwargs):

        super(RedisPool, self).__init__(*args, **kwargs)
        self._context = ContextVar(f'{self.__class__.__name__}_{uuid.uuid1().hex}', default=None)
        self._name = None

    @property
    def name(self):
        return self._name

    def _echo_pool_info(self):
        free = self.maxsize - self.size + self.freesize
        if free < Utils.math.ceil(self.maxsize / 3):
            Utils.log.warning(
                f'Redis pool not enough ({self._name}): (free:{free}/max:{self.maxsize})'
             )

    def get_context_client(self):
        ref = self._context.get()
        client = None
        if ref is not None:
            _client = ref() or None
            if _client and not _client.closed:
                client = _client
        return client

    async def get_client(self, alone=False):
        if self.closed:
            return
        self._echo_pool_info()
        client = Redis(await self.acquire())
        if not alone:
            self._context.set(weakref.ref(client))

        return client

    async def release(self, client: Redis):
        with catch_error():
            if self.closed:
                return

            client.close()
            await client.wait_closed()
            super().release(client._pool_or_conn)

    def get_context_manager_client(self, alone=False):

        return _AsyncContextManager(self, alone=alone)

    async def health(self):
        try:

            async with self.get_context_manager_client(alone=True) as client:

                await client.ping()
                return True

        except Exception:

            Utils.log.error(traceback.format_exc())

            return False


class MutexLock:
    """基于Redis实现的分布式锁，当await lock.acquire()一旦获取锁成功，会自动触发看门狗，持续对lock进行检查并保持持有
        直到跳出async with管理或者手动调用release()
        使用方法1（建议）：
            async with MutexLock(redis_pool, 'test-key', 60) as lock:
                is_locked = await lock.acquire()
                if is_locked:
                    # do something
                else:
                    # do something

        当你需要在程序生命周期内保持锁占有时，可使用方法2
        使用方法2：
            lock = MutexLock(redis_pool, 'test-key', 60)
            try:
                is_locked = await lock.acquire()
                if is_locked:
                    # do something
                else:
                    # do something
            except Exception as e:
                # do something
            finally:
                await lock.release()
    """

    _renew_script = '''
    if redis.call("get",KEYS[1]) == ARGV[1] and redis.call("ttl",KEYS[1]) > 0 then
        return redis.call("expire",KEYS[1],ARGV[2])
    else
        return 0
    end
    '''

    _unlock_script = '''
    if redis.call("get",KEYS[1]) == ARGV[1] then
        return redis.call("del",KEYS[1])
    else
        return 0
    end
    '''

    def __init__(self, redis_pool: RedisPool, key, expire=60):

        self._redis_pool = redis_pool
        assert expire > 3, '"expire" is too small'
        self._expire = expire
        self._key = key
        self._lock_tag = f'process_lock_{key}'
        self._lock_val = Utils.uuid.uuid1().hex.encode()

        self._locked = False
        self._cache = None
        self._err_count = 0

        # watch dog
        self._watcher = None

    async def __aenter__(self):

        await self._init_conn()

        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):

        await self.release()

    def __del__(self):
        if self._cache:
            asyncio.create_task(self.release())

    def __repr__(self):
        return f'{self.__class__.__name__}<{self._key}>'

    @property
    def locked(self):
        return self._locked

    async def _do_heartbeat(self):
        """watch dog continuous to refresh the lock"""
        res = await self.acquire(0)
        Utils.log.info(f'{str(self)} Watcher lock result: {res}')

    async def _init_conn(self, reset=False):
        """init redis conn or reset redis conn"""
        try:
            if self._cache is None or reset:
                if reset and self._cache is not None:
                    await self.release()
                self._cache = await self._redis_pool.get_client(alone=True)
                if not self._cache:
                    raise ValueError('can not get redis client')
                self.init_watcher()
                return True
        except Exception:
            Utils.log.error(f'{str(self)} init redis conn error')
            Utils.log.error(traceback.format_exc())

    def init_watcher(self):
        """init watch dog"""
        interval = Utils.math.floor(self._expire / 3)
        self._watcher = IntervalTask(interval or 3, self._do_heartbeat, tag=self.__class__.__name__)

    async def exists(self):
        """check the lock exist or not"""
        await self._init_conn()
        if await self._cache.exists(self._lock_tag):
            return True
        else:
            return False

    async def acquire(self, timeout=0):
        """acquire the lock"""
        assert not self._redis_pool.closed
        try:
            if self._locked:
                self._locked = await self.renew()
            else:
                await self._init_conn()

                params = {
                    r'key': self._lock_tag,
                    r'value': self._lock_val,
                    r'expire': self._expire,
                    r'exist': Redis.SET_IF_NOT_EXIST,
                }

                async for _ in AsyncCircularForTimeout(timeout):

                    if await self._cache.set(**params):
                        self._locked = True

                    if self._locked or timeout == 0:
                        break
            if self._locked and not self._watcher.running:
                self._watcher.start()

            if not self._locked:
                await self.release()

        except aioredis.errors.ConnectionClosedError:
            self._err_count += 1
            Utils.log.error(f'{str(self)} acquire conn closed error')
            if self._err_count >= 3:
                # If the connection is damaged, automatic reset to ensure service resume
                reset_res = await self._init_conn(reset=True)
                if reset_res:
                    self._err_count = 0

        return self._locked

    async def wait(self, timeout=0):

        async for _ in AsyncCircularForTimeout(timeout):

            if not await self.exists():
                return True
        else:
            return False

    async def renew(self):
        """renew the lock when lock held"""
        if self._cache is None:
            self._locked = False

        if self._locked:
            self._locked = False
            if await self._cache.eval(self._renew_script, [self._lock_tag], [self._lock_val, self._expire]):
                self._locked = True
            else:
                self._locked = False

        return self._locked

    async def release(self):
        """release the lock, and stop the watch dog"""
        with catch_error():
            if self._watcher and self._watcher.running:
                self._watcher.stop()
                self._watcher = None
            if self._locked and self._cache:
                await self._cache.eval(self._unlock_script, [self._lock_tag], [self._lock_val])
                self._locked = False

            if self._cache:
                await self._redis_pool.release(self._cache)
                self._cache = None


class FrequencyLimiter:
    DEFAULT_EXPIRE = 60

    def __init__(self, redis_pool: RedisPool, granularity, limit, conn_num=1, alone=False, ntp=None):

        self._redis_pool = redis_pool
        self._granularity = granularity
        self._limit = limit
        self._ntp = ntp
        self._alone = alone
        self._conn_num = conn_num
        self._expire = max([self._granularity, self.DEFAULT_EXPIRE])
        self._clients = []
        self._lock = asyncio.Lock()
        asyncio.create_task(self._init_clients())

    async def _init_clients(self, reset=False):
        async with self._lock:
            if reset:
                for client in self._clients:
                    await self._redis_pool.release(client)
                self._clients = []

            if not self._clients:
                self._clients = await asyncio.gather(
                    *[self._redis_pool.get_client(alone=self._alone) for _ in range(self._conn_num)])

            return self._clients[int(time.time() * 1000) % self._conn_num]

    async def get_client(self):

        if not self._clients:
            await self._init_clients()

        return self._clients[int(time.time() * 1000) % len(self._clients)]

    def generate_key(self, key):

        if self._ntp:
            divide = int(self._ntp.timestamp) // self._granularity
        else:
            divide = int(time.time()) // self._granularity

        return f'{self.__class__.__name__}_{key}_{divide}'

    async def incrby(self, key, incr_num=1):

        res = 0
        for _ in range(3):
            client = await self.get_client()

            try:

                pipeline = client.pipeline()
                _key = self.generate_key(key)

                pipeline.incrby(_key, incr_num)
                pipeline.expire(_key, self._expire)

                res, _ = await pipeline.execute()
                break
            except aioredis.errors.ConnectionClosedError:
                await self._init_clients(reset=True)
                continue

        return res

    async def is_limited(self, key, incr_num=1):

        res = await self.incrby(key, incr_num)
        if res > self._limit:
            return True
        return False

    async def release(self):
        if self._clients:
            for client in self._clients:
                await self._redis_pool.release(client)


class RedisPoolManager(SafeSingleton, BasePoolManager):
    """Redis连接池管理器"""
    def __init__(self):
        super().__init__()
        self._event_bus_pool = {}

    def __await__(self):

        if not self._pools:

            for pool_name, pool_config in self._config.items():
                pool_config.setdefault('pool_cls', RedisPool)
                self._pools[pool_name] = yield from aioredis.create_pool(**pool_config).__await__()
                setattr(self._pools[pool_name], '_name', pool_name)

                Utils.log.success(f'redis pool {self.pool_status(pool_name)} initialized')

            self._ready = True

        return self

    def close(self):
        if not self._pools:
            return
        pools, self._pools = self._pools, {}
        for pool in pools.values():
            pool.close()

    def pool_status(self, pool_name='default'):
        pool = self._pools[pool_name]
        return f'<[{pool_name}] {pool.address} [db:{pool.db}, size:[{pool.minsize}:{pool.maxsize}], free:{pool.freesize}]>'

    def allocate_lock(self, key, expire=60, pool_name='default'):
        assert self._ready, f'{self.__class__.__name__} does not initialize'

        return MutexLock(self._pools[pool_name], key, expire=expire)

    def allocate_limiter(self, pool_name='default', granularity=1, limit=10, conn_num=1, alone=False, ntp=None):
        assert self._ready, f'{self.__class__.__name__} does not initialize'

        return FrequencyLimiter(
            self._pools[pool_name],
            granularity=granularity,
            alone=alone,
            conn_num=conn_num,
            limit=limit,
            ntp=ntp
        )

    def allocate_event_bus(self, pool_name='default', channel_name='default'):
        assert self._ready, f'{self.__class__.__name__} does not initialize'
        assert pool_name in self._pools
        key = f'{pool_name}_|x|x|_{channel_name}'
        if key not in self._event_bus_pool:
            event_bus = RedisEventBus(self._pools[pool_name], channel_name)
            self._event_bus_pool[key] = event_bus

        return self._event_bus_pool[key]


redis_pool_manager = RedisPoolManager()


class RedisEventBus:
    """redis发布|订阅"""
    def __init__(self, redis_pool: RedisPool, channel_name):
        self._redis_pool = redis_pool

        self._channel = f'EVENT_BUS_[{self._redis_pool.name}]_[{channel_name}]'
        self._callbacks = {}

        asyncio.create_task(self._subscribe())

    def add_listener(self, tag: str, callback):
        assert asyncio.iscoroutinefunction(callback), 'callback must be coroutinefunction'
        if tag in self._callbacks:
            self._callbacks[tag].append(callback)
        else:
            self._callbacks[tag] = [callback]

    async def _subscribe(self):
        async for _ in AsyncCircularForSecond():
            if self._redis_pool.closed:
                continue
            async with self._redis_pool.get_context_manager_client(alone=True) as client:

                receiver, = await client.subscribe(self._channel)
                Utils.log.info(f'{self.__class__.__name__} channel({self._channel}) subscribed')

                async for message in receiver.iter():
                    asyncio.create_task(self._broadcast(message))

    async def _broadcast(self, message):

        message_dic = json.loads(message)
        tag = message_dic['tag']
        args = message_dic['args']
        kwargs = message_dic['kwargs']
        if tag in self._callbacks:
            await asyncio.gather(*[callback(*args, **kwargs) for callback in self._callbacks[tag]])

    async def publish(self, tag, *args, **kwargs):

        message = json.dumps(
            {
                'tag': tag,
                'args': list(args),
                'kwargs': kwargs
            }
        )

        async with self._redis_pool.get_context_manager_client() as client:
            await client.publish(self._channel, message)

        Utils.log.info(f'{self._channel } publish message args:{args} kwargs:{kwargs}')







class Node:

    def __init__(self, _key, _val, _prev=None, _next=None):
        self.key = _key
        self.val = _val
        self.prev = _prev
        self.next = _next

    def __str__(self):
        return f'Node({self.key}, {self.val})'



# -*- coding: utf-8 -*-

import functools
import traceback
import aiohttp

from wintersweet.utils.base import Utils
from wintersweet.asyncs.http.default import DEFAULT_SSL_CONTEXT, DEFAULT_DOWNLOAD_TIMEOUT, get_session_args
from wintersweet.asyncs.http.utils import _build_form_data, _file_sender
from wintersweet.asyncs.http.response import HTTPResponse, DownloadResponse, http_response_partial


async def request_ws(url, handle_response_class, **kwargs):
    resp = None
    try:
        session_args = await get_session_args()
        async with aiohttp.ClientSession(**session_args) as session:

            kwargs.setdefault('ssl', DEFAULT_SSL_CONTEXT)

            async with session.ws_connect(method=aiohttp.hdrs.METH_GET, url=url, **kwargs) as response:

                resp = await handle_response_class(response)

            Utils.log.info(f'ws {url} closed')

    except Exception:

        Utils.log.error(traceback.format_exc())

    return resp


async def request(method, url, handle_response_class=http_response_partial, **kwargs):
    """
    Usage::

      >>> from wintersweet.asyncs.http import api
      >>> req = await api.request('GET', 'https://httpbin.org/get')
      >>> req
      <HTTPResponse [200]>
    """
    resp = None
    try:
        session_args = await get_session_args()

        async with aiohttp.ClientSession(**session_args) as session:

            kwargs.setdefault('ssl', DEFAULT_SSL_CONTEXT)

            async with session.request(method=method, url=url, **kwargs) as response:

                resp = await handle_response_class(response)

            Utils.log.info(f'{method.upper()} {url} => status:{resp.status_code}')

    except aiohttp.ClientResponseError as err:

        Utils.log.error(f"{method.upper()} {url} status:{err.status}")

    except Exception:
        
        Utils.log.error(traceback.format_exc())

    return resp


async def get(url, params=None, **kwargs) -> HTTPResponse:
    """便捷的aiohttp GET请求，kwargs兼容aiohttp ClientSession.request参数"""
    return await request(aiohttp.hdrs.METH_GET, url, params=params, **kwargs)


async def options(url, **kwargs) -> HTTPResponse:
    """便捷的aiohttp OPTIONS请求，kwargs兼容aiohttp ClientSession.request参数"""
    return await request(aiohttp.hdrs.METH_OPTIONS, url, **kwargs)


async def head(url, **kwargs) -> HTTPResponse:
    """便捷的aiohttp HEAD请求，kwargs兼容aiohttp ClientSession.request参数"""
    kwargs.setdefault('allow_redirects', False)
    return await request(aiohttp.hdrs.METH_HEAD, url, **kwargs)


async def post(url, data=None, files=None, json=None, **kwargs) -> HTTPResponse:
    """便捷的aiohttp POST请求，kwargs兼容aiohttp ClientSession.request参数"""
    if files:
        # [(name, (filename, open(file_path, 'rb'))), ....]

        data = _build_form_data(data, files)
    kwargs['data'] = data
    return await request(aiohttp.hdrs.METH_POST, url, json=json, **kwargs)


async def put(url, data=None, files=None, json=None, **kwargs) -> HTTPResponse:
    """便捷的aiohttp PUT请求，kwargs兼容aiohttp ClientSession.request参数"""
    if files:
        # [(name, (filename, open(file_path, 'rb'))), ....]

        data = _build_form_data(data, files)
    kwargs['data'] = data

    return await request(aiohttp.hdrs.METH_PUT, url, json=json, **kwargs)


async def patch(url, data=None, files=None, json=None, **kwargs) -> HTTPResponse:
    """便捷的aiohttp PATCH请求，kwargs兼容aiohttp ClientSession.request参数"""
    if files:
        # [(name, (filename, open(file_path, 'rb'))), ....]

        data = _build_form_data(data, files)
    kwargs['data'] = data
    return await request(aiohttp.hdrs.METH_PATCH, url, json=json, **kwargs)


async def delete(url, **kwargs) -> HTTPResponse:
    """便捷的aiohttp DELETE请求，kwargs兼容aiohttp ClientSession.request参数"""
    return await request(aiohttp.hdrs.METH_DELETE, url, **kwargs)


async def request_stream(url, file_path, method=aiohttp.hdrs.METH_POST, **kwargs):
    """流式传输大文件"""
    return await request(method, url=url, data=_file_sender(file_path), **kwargs)


async def download(url, file_path, method=aiohttp.hdrs.METH_GET, **kwargs) -> DownloadResponse:
    """文件下载"""
    kwargs.setdefault('timeout', DEFAULT_DOWNLOAD_TIMEOUT)

    return await request(
        method=method,
        url=url,
        handle_response_class=functools.partial(DownloadResponse, file_path),
        **kwargs
    )





class TaskInterface:

    def __init__(self):
        self._running = False

    @property
    def running(self):
        return self._running

    def start(self):
        raise InterruptedError()

    def stop(self):
        raise InterruptedError()





import bson
import datetime
from motor.motor_asyncio import AsyncIOMotorClient

from wintersweet.asyncs.pool.base import BasePoolManager
from wintersweet.utils.base import Utils
from wintersweet.utils.metaclass import SafeSingleton


class MongoPool(AsyncIOMotorClient):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._name = None

    @property
    def name(self):
        return self._name

    async def health(self):
        result = False

        try:
            result = bool(await self.server_info())
        except Exception as err:
            Utils.log.error(err)

        return result


class MongoPoolManager(SafeSingleton, BasePoolManager):
    """mongo池化管理器"""
    def __init__(self):
        super(MongoPoolManager, self).__init__()

    def __await__(self):
        if self._ready:
            return self

        for pool_name, pool_config in self._config.items():
            yield
            self._pools[pool_name] = MongoPool(**pool_config)
            setattr(self._pools[pool_name], '_name', pool_name)
            Utils.log.success(
                f"mongo pool {self.pool_status(pool_name)} initialized"
            )

        self._ready = True

    def pool_status(self, pool_name='default'):
        pool = self._pools[pool_name]

        return f'<[{pool_name}] {pool.address} [size:[{pool.min_pool_size}:{pool.max_pool_size}]]>'

    def _echo_pool_info(self, pool_name='default'):
        pool = self._pools[pool_name]
        for address, server in pool.delegate._topology._servers.items():

            pool_size = len(server.pool.sockets) + server.pool.active_sockets

            if (pool.max_pool_size - pool_size) < Utils.math.ceil(pool.max_pool_size / 3):
                Utils.log.warning(
                    f'Mongo connection pool not enough ({pool_name}){address}: '
                    f'{pool_size}/{pool.max_pool_size}'
                )

    def get_client(self, pool_name='default'):
        assert pool_name in self._pools, f'pool_name: "{pool_name}" not found'
        self._echo_pool_info(pool_name)
        pool = self._pools[pool_name]
        return pool

    def get_collection(self, db_name, collection_name, pool_name='default'):
        """获取指定集合，无须集合存在，操作时将自动创建集合"""
        return self.get_client(pool_name)[db_name][collection_name]

    def get_collection_by_date(self, db_name, collection_name, date: datetime.date = None, pool_name='default'):
        """按日期切分集合，无须集合存在，操作时将自动创建集合"""
        if not date:
            date = datetime.datetime.now()
        date_str = Utils.datetime2time(date, format_type='%Y%m%d')
        return self.get_client(pool_name)[db_name][f'{collection_name}_{date_str}']

    def get_collection_by_month(self, db_name, collection_name, date: datetime.date = None, pool_name='default'):
        """按月切分集合，无须集合存在，操作时将自动创建集合"""
        if not date:
            date = datetime.datetime.now()
        month_str = Utils.datetime2time(date, format_type='%Y%m')
        return self.get_client(pool_name)[db_name][f'{collection_name}_{month_str}']

    @staticmethod
    def convert_mongo_id(mongo_id):
        """转换字符串mongo id为bson ObjectId"""
        try:
            return bson.ObjectId(mongo_id)
        except bson.errors.InvalidId:
            return bson.ObjectId()

    def close(self):
        _pools, self._pools = self._pools, {}
        for pool in _pools:
            pool.close()


mongo_pool_manager = MongoPoolManager()



import time

from wintersweet.utils.structure import Node


class TimestampNode(Node):

    def __init__(self, *args, **kwargs):
        super(TimestampNode, self).__init__(*args, **kwargs)
        self.timestamp = time.time()


class LRUCache:
    """
    使用频率排列缓存，最近使用过的靠前，满水位时优先剔除过期缓存，再剔除最长时间未使用缓存
    """
    def __init__(self, maximum: int = 100, ttl=3600):

        self._ttl = ttl
        self._maximum = maximum
        self._node_cache = {}
        self._size = 0

        self._head = None
        self._tail = None

    def __delete(self, node):

        self._size -= 1
        if node == self._head:
            temp_node = node.next
            if temp_node:
                temp_node.prev = None
                node.next = None
            self._head = temp_node

        elif node == self._tail:
            temp_node = node.prev
            if temp_node:
                temp_node.next = None
                node.prev = None
            self._tail = temp_node

        else:
            node.prev.next = node.next
            node.next.prev = node.prev

    def __append(self, node):
        self._size += 1
        if self._head is None:
            self._head = self._tail = node
        else:
            node.next = self._head
            self._head.prev = node

            self._head = node

    def __timeout_remove(self):
        temp = self._tail
        while temp and self._size >= self._maximum:
            prev = temp.prev
            if temp.timestamp + self._ttl < time.time():
                self.__delete(temp)
                self._node_cache.pop(temp.key)
            temp = prev

        while self._size >= self._maximum:
            tail_key = self._tail.key
            self.__delete(self._tail)
            self._node_cache.pop(tail_key)

    def get(self, key):

        if key not in self._node_cache:
            return None

        node = self._node_cache[key]
        if node.timestamp + self._ttl < time.time():
            self.__delete(node)
            self._node_cache.pop(key)

            return None

        self.__delete(node)
        self.__append(node)

        return node.val

    def put(self, key, val):

        if key in self._node_cache:
            node = self._node_cache[key]
            node.val = val

            self.__delete(node)
            self.__append(node)
        else:
            self.__timeout_remove()

            node = TimestampNode(key, val)
            self.__append(node)
            self._node_cache[key] = node

    def top(self):
        return self._head.val

    def list(self):

        node_list = []
        if self._head is None:
            return node_list

        temp = self._head
        while temp is not None:
            if temp.timestamp + self._ttl < time.time():
                temp = temp.next
                continue
            node_list.append((temp.key, temp.val))
            temp = temp.next

        return node_list








from fastapi import HTTPException
from fastapi.encoders import jsonable_encoder
from fastapi.exceptions import RequestValidationError
from starlette.requests import Request

from wintersweet.framework.fastapi.response import Response, HTTP400Response, HTTPExceptionResponse
from wintersweet.utils.base import Utils


async def http_exception_handler(
        request: Request, exc: HTTPException
) -> Response:

    headers = getattr(exc, 'headers', None)
    if headers:
        return HTTPExceptionResponse(status_code=exc.status_code, detail=exc.detail, headers=exc.headers)
    else:
        return HTTPExceptionResponse(status_code=exc.status_code, detail=exc.detail)


async def request_validation_exception_handler(
        request: Request, exc: RequestValidationError
) -> Response:
    Utils.log.error(f'{exc.body} --> {[err.get("loc") for err in exc.errors()]}')

    return HTTP400Response(content=jsonable_encoder(exc.errors()))





import asyncio
import traceback

from typing import Dict
from starlette.websockets import WebSocket, WebSocketDisconnect, WebSocketState

from wintersweet.utils.base import Utils


class WsChatConnectionManager:
    """ws会话管理器，超时无回应自动断开
    Example：
    >>>manager = WsChatConnectionManager()
    >>>router = APIRouter()

    >>>@router.websocket("/ws/{user_id}")
    >>>async def websocket_endpoint(ws: WebSocket, user_id: int=Query(...)):
    >>>     await manager.append(user_id, ws, MessageHandler)
    >>>     await manager.handle_message(user_id)
    """
    def __init__(self, limit=None, heartbeat_timeout=10, timeout_count=3):
        self._limit = limit
        self._heartbeat_timeout = heartbeat_timeout
        self._connections: Dict[str, WebSocket] = {}
        self._message_handlers = {}
        self._timeout_count = timeout_count
        self._lock = asyncio.Lock()

    async def append(self, key, ws, message_handle_class):
        """加入会话"""
        if self._limit and len(self._connections) >= self._limit:
            raise RuntimeWarning(f'ws connections out of limit, {key} append failed!')

        assert hasattr(message_handle_class, 'on_message') and hasattr(message_handle_class, 'hello')

        async with self._lock:

            if key in self._connections:
                await self._connections[key].close()

            self._connections[key] = ws
            self._message_handlers[key] = message_handle_class

    async def handle_message(self, key, receive_func='receive_text'):
        """处理会话循环"""
        try:
            ws: WebSocket = self._connections[key]
            await ws.accept()

            if not hasattr(ws, receive_func):
                await ws.close()
                raise AttributeError(f'Attribute "{receive_func}" not found')

            receive = getattr(ws, receive_func)
            message_handler = self._message_handlers[key]
            count = 0
            await message_handler.hello(ws, count)
            while True:
                try:
                    message = await asyncio.wait_for(receive(), self._heartbeat_timeout)
                except asyncio.TimeoutError:
                    count += 1
                    if count >= self._timeout_count:
                        await ws.close()

                        raise WebSocketDisconnect()

                    await message_handler.hello(key, ws, count)
                    continue

                count = 0
                await message_handler.on_message(key, ws, message)

        except WebSocketDisconnect:
            self._connections.pop(key)
            self._message_handlers.pop(key)
            Utils.log.info(f'ws:[{key}] disconnected')

        except Exception:
            Utils.log.error(traceback.format_exc())

    def get_client(self, key):
        return self._connections.get(key)

    async def send_message(self, key, message):
        """根据key发送消息"""

        ws = self.get_client(key)
        if not ws or ws.client_state == WebSocketState.DISCONNECTED:
            Utils.log.warning(f'ws:[{key}] maybe disconnected, can not send message')
            return False

        try:

            if isinstance(message, bytes):
                await ws.send_bytes(message)
            elif isinstance(message, str):
                await ws.send_text(message)
            elif isinstance(message, (list, dict)):
                await ws.send_json(message)
            else:
                await ws.send(message)

            return True

        except WebSocketDisconnect:
            Utils.log.error(f'ws:[{key}] disconnected, send message failed')

        except Exception:
            Utils.log.error(traceback.format_exc())

        return False


class MessageHandler:

    @staticmethod
    async def hello(key, ws, count):

        pass

    @staticmethod
    async def on_message(key, ws, message):

        pass




__author__ = r'wsb310@gmail.com'  # copy

import fastapi
from starlette.concurrency import run_in_threadpool
from starlette.routing import iscoroutinefunction_or_partial

from starlette.types import Receive, Scope, Send, ASGIApp

from wintersweet.framework.fastapi.middlewares import RequestIDMiddleware
from wintersweet.framework.fastapi.response import Response, ErrResponse
from wintersweet.utils.base import Utils


class APIRoute(fastapi.routing.APIRoute):

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

        self.app = self.request_response()

    async def prepare(self, request):

        pass

    def request_response(self) -> ASGIApp:

        func = self.get_route_handler()

        is_coroutine = iscoroutinefunction_or_partial(func)

        async def app(scope: Scope, receive: Receive, send: Send) -> None:
            request = Request(scope, receive, send, self)

            try:

                await self.prepare(request)

                if is_coroutine:
                    response = await func(request)
                else:
                    response = await run_in_threadpool(func, request)

                await response(scope, receive, send)

            except ErrResponse as err:

                Utils.log.warning(f'ErrResponse: {request.path}\n{err}\n{request.debug_info}')

                await err(scope, receive, send)

        return app


class Request(fastapi.Request):

    def __init__(self, scope: Scope, receive: Receive, send: Send, api_route: APIRoute):

        super().__init__(scope, receive, send)

        self._api_route = api_route

        self._request_id = RequestIDMiddleware.get_request_id()

    @property
    def request_id(self):

        return self._request_id

    @property
    def debug_info(self):

        return {
            'request_id': self._request_id,
            'type': self.scope.get(r'type'),
            'http_version': self.scope.get(r'http_version'),
            'server': self.scope.get(r'server'),
            'client': self.scope.get(r'client'),
            'scheme': self.scope.get(r'scheme'),
            'method': self.scope.get(r'method'),
            'root_path': self.scope.get(r'root_path'),
            'path': self.scope.get(r'path'),
            'query_string': self.scope.get(r'query_string'),
            'headers': self.scope.get(r'headers'),
        }

    @property
    def route(self):

        return self._api_route

    @property
    def path(self):

        return self._api_route.path

    @property
    def tags(self):

        return self._api_route.tags

    @property
    def referer(self):

        return self.headers.get(r'Referer')

    @property
    def client_ip(self):

        return self.x_forwarded_for or self.client_host

    @property
    def client_host(self):

        return self.headers.get(r'X-Real-IP', self.client.host)

    @property
    def x_forwarded_for(self):

        return self.headers.get(r'X-Forwarded-For', r'').split(',')[0]

    @property
    def content_type(self):

        return self.headers.get(r'Content-Type')

    @property
    def content_length(self):

        result = self.headers.get(r'Content-Length', r'')

        return int(result) if result.isdigit() else 0

    def get_header(self, name, default=None):

        return self.headers.get(name, default)


class APIRouter(fastapi.APIRouter):

    def __init__(
            self,
            *,
            prefix=r'',
            default_response_class=Response,
            route_class=APIRoute,
            **kwargs
    ):

        super().__init__(
            prefix=prefix,
            default_response_class=default_response_class,
            route_class=route_class,
            **kwargs
        )


import logging
import pytz

from apscheduler.schedulers.asyncio import AsyncIOScheduler
from apscheduler.triggers.cron import CronTrigger

from wintersweet.asyncs.task.interfaces import TaskInterface
from wintersweet.utils.base import Utils

TIMEZONE = pytz.timezone(Utils.os.getenv('_TIMEZONE', r'Asia/Shanghai'))


logging.getLogger(r'apscheduler').setLevel(logging.WARNING)


class BaseAsyncIOScheduler(AsyncIOScheduler):
    def __init__(self, tag, **kwargs):
        self._tag = tag
        super().__init__(**kwargs)

    @property
    def tag(self):
        return self._tag


class AsyncIOSchedulerManager:
    """线程安全的定时任务管理器"""

    _schedulers = {}

    @classmethod
    def create_scheduler(cls, tag='default'):

        if tag in cls._schedulers:
            return cls._schedulers[tag]

        scheduler = BaseAsyncIOScheduler(
            tag=tag,
            job_defaults={
                r'coalesce': False,
                r'max_instances': 1,
                r'misfire_grace_time': 10
            },
            timezone=TIMEZONE
        )
        result = cls._schedulers.setdefault(tag, scheduler)
        if id(result) == id(scheduler):
            result.start()

        return result

    @classmethod
    def stop(cls):
        _schedulers, cls._schedulers = cls._schedulers, {}
        for scheduler in _schedulers.values():
            scheduler.shutdown()

        cls._schedulers.clear()


class BaseTask(TaskInterface):

    def __init__(self, func, tag='default', *args, **kwargs):
        super(BaseTask, self).__init__()
        self._scheduler = None
        self._tag = tag
        self._func = Utils.package_task(func, *args, **kwargs)
        self._job = None

    @property
    def scheduler(self):
        return self._scheduler

    def start(self):
        if not self._job:
            self._scheduler = AsyncIOSchedulerManager.create_scheduler(self._tag)
            self._job = self._scheduler.add_job(**self._job_args())
            self._running = True
        return self._job

    def stop(self):
        assert self._job is not None, f'{self.__class__.__name__} has no start'
        self._scheduler.remove_job(self._job.id)
        self._job = None
        self._running = False
        return True

    def _job_args(self):

        raise InterruptedError


class IntervalTask(BaseTask):
    """间隔触发任务"""

    def __init__(self, interval: int, func, tag='IntervalTask', *args, **kwargs):
        super(IntervalTask, self).__init__(func, tag=tag, *args, **kwargs)
        self._interval = interval

    def _job_args(self):
        return {
            'func': self._func,
            'trigger': r'interval',
            'seconds': self._interval
        }


class CrontabTask(BaseTask):
    """定时触发任务"""
    def __init__(self, crontab, func, tag='CrontabTask', *args, **kwargs):
        super(CrontabTask, self).__init__(func, tag, *args, **kwargs)
        self._crontab = crontab

    def _job_args(self):
        return {
            'func': self._func,
            'trigger': CronTrigger.from_crontab(self._crontab, TIMEZONE),
        }



from wintersweet.utils.base import Utils


class _AsyncCircular:
    def __init__(self, max_times=None):

        self._current = 0
        self._max_times = max_times

    def __aiter__(self):

        return self

    async def __anext__(self):

        if self._current > 0:

            if self._is_stop():
                raise StopAsyncIteration()

            await self._sleep()

        self._current += 1

        return self._current - 1

    def _is_stop(self):
        return self._max_times and (self._max_times > 0) and (self._max_times <= self._current)

    async def _sleep(self):

        raise InterruptedError()


class AsyncCircularForFrame(_AsyncCircular):
    """帧异步循环器

    提供一个循环体内的代码重复执行管理逻辑，可控制循环次数、执行间隔(LoopFrame)

    async for index in AsyncCircularForFrame():
        pass

    其中index为执行次数，从0开始

    """

    def __init__(self, max_times, interval=0xfff):
        """
        :param max_times: 循环次数
        :param interval: 执行间隔（帧）
        """
        super(AsyncCircularForFrame, self).__init__(max_times)

        self._interval = interval

    async def _sleep(self):
        await Utils.wait_frame(self._interval)


class AsyncCircularForTimeout(_AsyncCircular):
    """异步超时循环器，在超时时间内，以interval为循环间隔"""
    def __init__(self, timeout, interval=0xfff):
        """
        :param timeout: 超时时长
        :param interval: 执行间隔（帧）
        """
        super(AsyncCircularForTimeout, self).__init__()

        self._interval = interval
        self._timeout_time = Utils.loop_time() + timeout

    def _is_stop(self):
        return self._timeout_time < Utils.loop_time()

    async def _sleep(self):
        await Utils.wait_frame(self._interval)


class AsyncCircularForSecond(_AsyncCircular):
    """秒异步循环器

    可控制循环次数、执行间隔(Second)

    async for index in AsyncCircularForSecond():
        pass

    其中index为执行次数，从0开始

    """
    def __init__(self, max_times=None, interval=1):

        super(AsyncCircularForSecond, self).__init__(max_times)
        self._interval = interval

    async def _sleep(self):
        await Utils.asyncio.sleep(self._interval)





from contextlib import contextmanager
from wintersweet.utils.base import Utils


@contextmanager
def catch_error():
    """异常捕获，打印error级日志
    Usage：
        with catch_error():
            ...
    通过with语句捕获异常
    """

    try:

        yield

    except Exception as err:

        Utils.log.exception(err)




import asyncio

import aioredis


class _AsyncContextManager:

    def __init__(self, pool, alone=False):

        self._pool = pool
        self._is_ctx = False
        self._alone = alone

    async def __aenter__(self):
        if not self._alone:
            self._client = self._pool.get_context_client()
            if not self._client:
                self._client = await self._pool.get_client()
            else:
                self._is_ctx = True
        else:
            self._client = await self._pool.get_client()

        return self._client

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self._is_ctx:
            return

        _pool, self._pool = self._pool, None
        _client, self._client = self._client, None

        if hasattr(_client, 'close'):
            res = _client.close()
            if asyncio.iscoroutine(res):
                await res

        if isinstance(_pool, aioredis.ConnectionsPool):
            await _pool.release(_client)



from enum import Enum
from fastapi.responses import UJSONResponse
from starlette.status import HTTP_400_BAD_REQUEST, HTTP_500_INTERNAL_SERVER_ERROR

from wintersweet.framework.fastapi.middlewares import RequestIDMiddleware


class Response(UJSONResponse):
    def __init__(self, content=None, code=0, status_code=200, extra=None, *args, **kwargs):
        if isinstance(code, Enum):
            code = code.value
        self._code = code
        self._msg = None
        try:
            self._msg = self.build_msg()
        except Exception:
            if code == 0:
                self._msg = 'Success'
            else:
                self._msg = 'Unknown Error'

        self._data = content
        self._extra = extra
        self._request_id = RequestIDMiddleware.get_request_id()

        super().__init__(content=content, status_code=status_code, *args, **kwargs)

    def build_msg(self):
        raise InterruptedError()

    @property
    def data(self):
        return self._data

    def render(self, content):
        return super().render(
            dict(code=self._code, data=content, msg=self._msg, request_id=self._request_id, extra=self._extra)
        )

    def __bool__(self):

        return self._code == 0


class HTTP404Response(Response):
    def __init__(self):
        super(HTTP404Response, self).__init__(code=404, status_code=404)

    def build_msg(self):
        return r'Not Found'


class HTTP405Response(Response):
    def __init__(self):
        super(HTTP405Response, self).__init__(code=405, status_code=405)

    def build_msg(self):
        return r'Method Not Allowed'


class HTTP400Response(Response):
    def __init__(self, content=None, headers=None):
        super(HTTP400Response, self).__init__(
            content=content,
            code=400,
            status_code=HTTP_400_BAD_REQUEST,
            headers=headers
        )

    def build_msg(self):
        return r'Invalid Arguments'


class HTTP500Response(Response):
    def __init__(self, content=None, headers=None):
        super(HTTP500Response, self).__init__(
            content=content,
            code=500,
            status_code=HTTP_500_INTERNAL_SERVER_ERROR,
            headers=headers
        )

    def build_msg(self):
        return r'Service Unavailable'


class HTTPExceptionResponse(Response):
    def __init__(self, status_code=503, detail='Service Unavailable', headers=None):
        super(HTTPExceptionResponse, self).__init__(
            code=status_code,
            status_code=status_code,
            headers=headers
        )
        self.detail = detail

    def build_msg(self):
        return self.detail


class ErrResponse(Response, Exception):
    pass



# -*- coding: utf-8 -*-

import setuptools

from wintersweet import __version__


with open(r'README.md', r'r', encoding=r'utf8') as stream:
    long_description = stream.read()

setuptools.setup(
    name=r'wintersweet',
    version=__version__,
    license=r'Apache License Version 2.0',
    platforms=[r'all'],
    author=r'Xixi.Dong',
    author_email=r'xq1889@163.com',
    description=r'Network Development Suite',
    long_description=long_description,
    long_description_content_type=r'text/markdown',
    url=r'https://gitee.com/xixigroup/wintersweet',
    packages=setuptools.find_packages(),
    python_requires=r'>= 3.8',
    install_requires=[
        r'APScheduler==3.7.0',
        r'Pillow==8.2.0',
        r'PyJWT==2.1.0',
        r'PyYAML==5.4.1',
        r'SQLAlchemy==1.3.24',
        r'aiohttp==3.7.4',
        r'aiomysql==0.0.21',
        r'aioredis==1.3.1',
        r'async-timeout==3.0.1',
        r'cachetools==4.2.2',
        r'cryptography==3.4.7',
        r'fastapi==0.65.1',
        r'gunicorn==20.1.0',
        r'hiredis==2.0.0',
        r'httptools==0.2.0',
        r'loguru==0.5.3',
        r'motor==2.4.0',
        r'ntplib==0.3.4',
        r'numpy==1.20.3',
        r'protobuf==3.17.1',
        r'psutil==5.8.0',
        r'pyahocorasick==1.4.2',
        r'pytest-asyncio==0.15.1',
        r'pytest==6.2.4',
        r'python-multipart==0.0.5',
        r'qrcode==6.1',
        r'ujson==4.0.2',
        r'uvicorn==0.13.4',
        r'uvloop==0.15.3;sys_platform!="win32"',
        r'xmltodict==0.12.0',
        r'aiofiles==0.8.0',
        r'elasticsearch==8.1.2',
    ],
    classifiers=[
        r'Programming Language :: Python :: 3.8',
        r'License :: OSI Approved :: Apache Software License',
        r'Operating System :: POSIX :: Linux',
    ],
    include_package_data=True
)



import functools
import os
import traceback
from typing import Optional

import aiofiles
from aiohttp import ClientResponse
from aiohttp.typedefs import DEFAULT_JSON_DECODER, JSONDecoder

from wintersweet.utils.base import Utils


class BaseResponse:
    def __init__(self, client_response: ClientResponse):

        self._client_response = client_response
        self._parsed = False
        self._success = False
        self._status_code = None

    @property
    def response(self):
        return self._client_response

    @property
    def parsed(self):
        return self._parsed

    @property
    def success(self):
        return self._client_response.status < 400

    @property
    def status_code(self):
        return self._client_response.status

    @property
    def headers(self):
        return self._client_response.headers

    @property
    def raw_headers(self):
        return self._client_response.raw_headers

    @property
    def request(self):
        return self._client_response.request_info

    def __await__(self):

        raise NotImplementedError()

    def __repr__(self):
        return f'<{self.__class__.__name__} [{self.status_code}]>'


class HTTPResponse(BaseResponse):

    def __init__(self, client_response: ClientResponse):

        super().__init__(client_response)
        self._content = None
        self._encoding = None

    def __await__(self):
        if not self.parsed:
            self._content = yield from self._client_response.read().__await__()
            self._encoding = self._client_response.get_encoding()
            self._client_response.release()
            self._parsed = True

        return self

    @property
    def text(self):
        return self._content.decode(self._encoding)

    @property
    def content(self):
        return self._content

    def get_text(self, encoding: Optional[str] = None):
        return self._content.decode(encoding or self._encoding)

    def json(self, encoding: Optional[str] = None, loads: JSONDecoder = DEFAULT_JSON_DECODER):
        bytes_data = self._content.strip()

        return loads(bytes_data.decode(encoding or self._encoding))


class DownloadResponse(BaseResponse):

    def __init__(self, file_path, client_response: ClientResponse):
        super().__init__(client_response)
        self._file = file_path

    def __await__(self):
        if not self._parsed:

            w = yield from aiofiles.open(self._file, "wb").__await__()
            try:
                while True:
                    chunk = yield from self._client_response.content.read(65535).__await__()
                    if not chunk:
                        break
                    yield from w.write(chunk)
                self._parsed = True

            except BaseException:
                Utils.log.error(traceback.format_exc())

            finally:
                yield from w.close()

            if not self._parsed:
                os.remove(self._file)

        return self


http_response_partial = functools.partial(HTTPResponse)



import importlib
import os

from . import global_setting


class DynamicSetting:

    def initialize(self):

        settings_path = os.environ.get('WINTERSWEET_SETTINGS_MODULE', 'wintersweet.framework.global_setting')

        self.__dict__['_settings'] = _Setting(settings_path)

    def __getattr__(self, item):

        if self.__dict__.get('_settings') is None:
            self.initialize()

        if hasattr(self._settings, item):
            val = getattr(self._settings, item)
            self.__dict__[item] = val
        else:
            val = None
        return val

    def __setattr__(self, key, value):
        if key == '_settings':
            self.__dict__.clear()
        else:
            if self.__dict__.get('_settings') is None:
                self.initialize()

            setattr(self._settings, key, value)
            self.__dict__[key] = value


class _Setting:

    def __init__(self, module_or_path):

        for item in dir(global_setting):
            if item.isupper():
                setattr(self, item, getattr(global_setting, item))

        if isinstance(module_or_path, str):
            mod = importlib.import_module(module_or_path)
        else:
            mod = module_or_path
        for item in dir(mod):
            if not item.startswith('_') and item.isupper():
                setattr(self, item, getattr(mod, item))


settings = DynamicSetting()

